import re
import hashlib
from .generics import Rule


class RegexRule(Rule):
    def __init__(
        self, regex, only_primitive_values=False, only_selected_values=None, **params
    ):
        super().__init__(
            **{**{"only_primitive_values": only_primitive_values}, **params}
        )
        self.regex = re.compile(f"^({regex})$", re.IGNORECASE)
        self.string = regex
        if only_selected_values:
            self.type_rule = TypeRule(only_selected_values)
        elif only_primitive_values:
            self.type_rule = TypeRule((str, bool, float, int))
        else:
            self.type_rule = TypeRule(object)

    def load(self, parser):
        super().load(parser)
        self.type_rule.load(parser)

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.string) + ")"

    def unique_hex(self):
        return hashlib.md5((self.salt + self.regex.pattern).encode("utf-8")).hexdigest()

    def apply_rule(self, field, value):
        return self.type_rule.apply_rule(field, value) and (
            self.regex.search(field) is not None
        )


class BoolRule(Rule):
    def __init__(self, condition, **params):
        super().__init__(**params)
        self.condition = condition

    def load(self, parser):
        super().load(parser)

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.condition) + ")"

    def unique_hex(self):
        return hashlib.md5(
            (self.salt + str(self.condition)).encode("utf-8")
        ).hexdigest()

    def apply_rule(self, field, value):
        if callable(self.condition):
            return self.condition(field, value)
        elif isinstance(self.condition, bool):
            return self.condition
        else:
            raise AttributeError("Invalid type for provided condition")


class TypeRule(Rule):
    def __init__(self, accepted_types, **params):
        super().__init__(**params)
        self.accepted_types = accepted_types
        if not isinstance(self.accepted_types, (list, tuple)):
            self.accepted_types = [self.accepted_types]
        self.accepted_types = tuple(self.accepted_types)

    def load(self, parser):
        super().load(parser)

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.accepted_types) + ")"

    def unique_hex(self):
        return hashlib.md5(
            (self.salt + str(self.accepted_types)).encode("utf-8")
        ).hexdigest()

    def apply_rule(self, field, value):
        return isinstance(value.value, self.accepted_types)


class AndRule(Rule):
    def __init__(self, rules, **params):
        super().__init__(**params)
        self.rules = rules

    def load(self, parser):
        super().load(parser)

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.rules) + ")"

    def unique_hex(self):
        return hashlib.md5(
            (self.salt + "".join([str(rule.__hash__()) for rule in self.rules])).encode(
                "utf-8"
            )
        ).hexdigest()

    def apply_rule(self, field, value):
        is_true = True
        for rule in self.rules:
            is_true = rule.apply_rule(field, value)
            if not is_true:
                break
        return is_true


class OrRule(Rule):
    def __init__(self, rules, **params):
        super().__init__(**params)
        self.rules = rules

    def load(self, parser):
        super().load(parser)

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.rules) + ")"

    def unique_hex(self):
        return hashlib.md5(
            (self.salt + "".join([str(rule.__hash__()) for rule in self.rules])).encode(
                "utf-8"
            )
        ).hexdigest()

    def apply_rule(self, field, value):
        is_true = False
        for rule in self.rules:
            is_true = rule.apply_rule(field, value)
            if is_true:
                break
        return is_true



def is_child_concept(type, parent_type, ontology):
    return type == parent_type or parent_type in ontology["concepts"][type]["parents"]

def get_label_keys(type, ontology):
    parents = set(ontology["concepts"][type]["parents"])
    parents.add(type)
    parents = list(parents)
    parents.reverse()
    label_keys = {} # for guaranteed insertion order
    for parent in parents:
        for label_key in ontology["concepts"][parent]["labelKeys"]:
            if not label_key:
                continue
            label_keys[label_key] = None
    return list(label_keys.keys())



import asyncio
from fastapi import Request, Body
from fastapi.responses import JSONResponse, StreamingResponse
from pydantic import BaseModel
from typing import List, Optional, Literal, Any
from io import BytesIO
import zipfile
from abc import ABC, abstractmethod
import json
from pydantic import ValidationError
from fastapi.exceptions import RequestValidationError
from starlette.exceptions import HTTPException as StarletteHTTPException
import ast
import uuid
from functools import wraps
import traceback
from copy import copy
import logging
    
from octostar.client import make_client

MAX_ERROR_MESSAGE_BYTES = 256
MAX_ERROR_TRACEBACK_BYTES = 10240
DEFAULT_PROCESSOR_SUFFIX = "main"


class CommonParsers(object):
    async def parse_form_list(form_data: str, format=Literal["csv", "python"]) -> Optional[List[str]]:
        try:
            if not form_data:
                return None
            if format == "csv":
                return form_data.split(",")
            elif format == "python":
                return ast.literal_eval(form_data)
        except BaseException as e:
            raise ValidationError(f"Parsing failed in parse_from_list: {str(e.__class__)} {str(e)}")


class Route(ABC):
    def route(route, **route_kwargs):
        def decorator(func):
            @wraps(func)
            async def async_wrapper(*args, **kwargs):
                return await func(*args, **kwargs)

            @wraps(func)
            def sync_wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            func_name = func.__name__
            if asyncio.iscoroutinefunction(func):
                wrapper = async_wrapper
            else:
                wrapper = sync_wrapper
            func_name = func.__name__
            wrapper._route_kwargs = route_kwargs
            wrapper._route_funcname = func_name
            route.routed_funcs.append(wrapper)
            return wrapper

        return decorator

    def include(route):
        def decorator(func):
            func_name = func.__name__
            if getattr(route, func_name, None):
                raise ValueError(f"Cannot have two functions with the same name {func_name} within the same Route!")
            setattr(route, func_name, func)
            return func

        return decorator

    def __init__(self, app, router=None):
        self.app = app
        self._router = router
        self.routed_funcs = []
        self.define_routes()

    @abstractmethod
    def define_routes(self):
        pass

    @property
    def router(self):
        return self._router or self.app

    def include_routes(self, router_mapping=None):
        for func in self.routed_funcs:
            route_kwargs = copy(func._route_kwargs)
            router = self.router
            if route_kwargs.get("router"):
                router = route_kwargs["router"]
            elif router_mapping:
                router = router_mapping.get(func._route_funcname, self.router)
            else:
                router = self.router
            route_kwargs.pop("router", None)
            router.add_api_route(**route_kwargs, endpoint=func)
        return self


class OctostarRoute(Route):
    def __init__(self, app, tasks_routes, celery_executor=None, router=None):
        self.app = app
        self._router = router
        self.routed_funcs = []
        self.tasks_routes = tasks_routes
        self.celery_executor = celery_executor
        self.endpoints = {}
        self.define_routes()

    def register_route(self, op, octostar_task):
        self.endpoints[op.strip("/")] = octostar_task

    def define_routes(self):
        if self.celery_executor:

            @Route.route(self, path="/task-state/{task_id}")
            async def get_task_status(task_id: str) -> JSONResponse:
                task_status = await self.tasks_routes.get_task(task_id, pop=False)
                task_status = task_status.model_dump(mode="json")["data"]["task_state"]
                return JSONResponse(task_status)

            @Route.route(self, path="/task-result/{task_id}")
            async def get_task_result(task_id: str) -> JSONResponse:
                return_data = await self.tasks_routes.get_task(task_id, pop=True)
                return_data = return_data.model_dump(mode="json")["data"]["data"]
                return JSONResponse(return_data)

            @Route.route(self, path="/{op}", methods=["POST"])
            async def send_task(
                op: str,
                os_context: dict = Body(...),
                jwt: str = Body(...),
                params: dict = Body(dict()),
            ) -> str:
                """
                Any request coming from Octostar (e.g. manifest) should enter from here.
                """
                path_params = []
                op = op.split("/")
                if len(op) > 1:
                    path_params = op[1:]
                op = op[0]
                query_params = params
                client = make_client(jwt)
                if op not in self.endpoints.keys():
                    raise StarletteHTTPException(401, f"Route {op} is forbidden for NiFi.")
                task_id = await self.celery_executor.send_task(
                    self.endpoints[op], args=[os_context, client, query_params]
                )
                return task_id

        else:

            @Route.route(self, path="/{op}", methods=["POST"])
            async def call_task(
                op: str,
                os_context: dict = Body(...),
                jwt: str = Body(...),
                params: dict = Body(dict()),
            ) -> str:
                """
                Any request coming from Octostar (e.g. manifest) should enter from here.
                """
                path_params = []
                op = op.split("/")
                if len(op) > 1:
                    path_params = op[1:]
                op = op[0]
                query_params = params
                client = make_client(jwt)
                if op not in self.endpoints.keys():
                    raise StarletteHTTPException(401, f"Route {op} is forbidden for NiFi.")
                result = await self.endpoints[op](os_context, client, query_params)
                return result

    @staticmethod
    def octostar_task(celery_executor, *args, **opts):
        def decorator(func):
            if celery_executor:
                serialized_func = celery_executor.serialized_io(func)
                task_func = celery_executor.app.task(*args, **opts)(serialized_func)
            else:

                @wraps(func)
                def octostar_func(*args, **kwargs):
                    return func(None, *args, **kwargs)

                task_func = octostar_func
            return task_func

        return decorator


class CommonModels(object):
    class OKResponseModel(BaseModel):
        message: str = "OK"
        status: str = "success"

    class DataResponseModel(BaseModel):
        data: Any
        status: str = "success"

    class ZipResponseModel(object):
        class Element(object):
            def __init__(self, file: bytes, path: str, filename: str, json: dict):
                self.file = file
                self.path = path
                self.filename = filename
                self.json = json

        def __init__(
            self,
            elements: List[Element],
            paths: Optional[List[str]] = None,
            filename: Optional[str] = None,
            status: str = "success",
        ):
            self.elements = elements
            self.paths = paths
            self.status = status
            self.filename = filename

        def get(self):
            zip_buffer = BytesIO()
            with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zip_file:
                for element in self.elements:
                    file_name = f"{element.path}/{element.filename}"
                    zip_file.writestr(file_name, element.file)
                    json_name = f"{element.path}/{element.filename}.dict"
                    zip_file.writestr(json_name, json.dumps(element.json, indent=4))
                for path in self.paths:
                    zip_file.writestr(path + "/", "")
                zip_file.writestr("response.dict", json.dumps({"status": self.status}, indent=4))
            zip_buffer.seek(0)
            headers = {}
            headers["Content-Disposition"] = "attachment;"
            filename = self.filename or (str(uuid.uuid4()) + ".zip")
            headers["Content-Disposition"] += f' filename="{filename}"'
            return StreamingResponse(zip_buffer, media_type="application/zip", headers=headers)


class ErrorLogFilter(logging.Filter):
    def __init__(self, silenced_excs: dict[type, callable]):
        super().__init__()
        self.silenced_excs = silenced_excs

    def _silenced(self, exc_value) -> bool:
        for etype, cond in self.silenced_excs.items():
            if isinstance(exc_value, etype) and cond(exc_value):
                return True
        return False

    def filter(self, record: logging.LogRecord) -> bool:
        if record.exc_info:
            exc_type, exc_value, _ = record.exc_info
            if self._silenced(exc_value):
                return False

        msg = record.getMessage()
        for etype in self.silenced_excs:
            if etype.__name__ in msg:
                if self.silenced_excs[etype](None):
                    return False

        return True


class DefaultErrorRoute:
    DEFAULT_STATUS_CODE_MAPPINGS = {
        StarletteHTTPException: lambda exc: exc.status_code,
        NotImplementedError: lambda exc: 501,
        AssertionError: lambda exc: 400,
        ValidationError: lambda exc: 422,
        RequestValidationError: lambda exc: 422,
    }

    DEFAULT_SILENCED_EXCEPTIONS = {
        RequestValidationError: lambda exc: True,
        ValidationError: lambda exc: True,
        StarletteHTTPException: lambda exc: exc.status_code in [400, 429, 422]
    }

    error_responses = {
        500: {
            "description": "Generic Server Error",
            "content": {
                "application/json": {"example": {"message": "Something unexpected went wrong!", "status": "error"}}
            },
        },
        501: {
            "description": "Not Implemented",
            "content": {
                "application/json": {
                    "example": {"message": "This method (with these parameters) is not implemented!", "status": "error"}
                }
            },
        },
        400: {
            "description": "Bad Request",
            "content": {
                "application/json": {
                    "example": {"message": "These two lists must have the same length!", "status": "error"}
                }
            },
        },
        422: {
            "description": "Validation Error",
            "content": {
                "application/json": {"example": {"message": "Validation error on 0 -> filename!", "status": "error"}}
            },
        },
    }

    def format_error(exc, body=b"", debug=False, excs_to_status_codes=DEFAULT_STATUS_CODE_MAPPINGS):
        """Generic Error Handler"""
        status_code = 500
        for exc_type, handler in excs_to_status_codes.items():
            if isinstance(exc, exc_type):
                status_code = handler(exc)
                break
        try:
            message = exc.message
        except:
            message = str(exc)
        if debug:
            message += "\n" + str(body)
        if len(message) > MAX_ERROR_MESSAGE_BYTES:
            message = message[-MAX_ERROR_MESSAGE_BYTES:]
        try:
            tcbk = "\n".join(traceback.format_exception(exc))
            if len(tcbk) > MAX_ERROR_TRACEBACK_BYTES:
                tcbk = tcbk[-MAX_ERROR_TRACEBACK_BYTES:]
        except:
            tcbk = None
        response_content = {"message": message, "status": "error"}
        if debug:
            response_content["traceback"] = tcbk
        return JSONResponse(status_code=status_code, content=response_content)

    async def handle_error(body: bytes, exc: Exception, debug: bool, excs_to_status_codes: dict):
        return DefaultErrorRoute.format_error(exc, body, debug, excs_to_status_codes)

    def add_default_exceptions_handler(
            fs_app,
            debug=False,
            excs_to_status_codes=None,
            silenced_excs=None,
    ):
        if excs_to_status_codes is None:
            excs_to_status_codes = DefaultErrorRoute.DEFAULT_STATUS_CODE_MAPPINGS
        if silenced_excs is None:
            silenced_excs = DefaultErrorRoute.DEFAULT_SILENCED_EXCEPTIONS

        async def _async_handle_error(request: Request, exc: Exception):
            return await DefaultErrorRoute.handle_error(b"", exc, debug, excs_to_status_codes)

        # Added all three since FastAPI seems to intercept some exceptions before Exception
        fs_app.add_exception_handler(RequestValidationError, _async_handle_error)
        fs_app.add_exception_handler(StarletteHTTPException, _async_handle_error)
        fs_app.add_exception_handler(Exception, _async_handle_error)

        log_filter = ErrorLogFilter(silenced_excs)
        for name in ("uvicorn.error", "uvicorn.access", "fastapi", "uvicorn"):
            logging.getLogger(name).addFilter(log_filter)

        logging.getLogger().addFilter(log_filter)

class RequestCancelledMiddleware:
    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return
        queue = asyncio.Queue()

        async def message_poller(sentinel, handler_task):
            nonlocal queue
            while True:
                message = await receive()
                if message["type"] == "http.disconnect":
                    handler_task.cancel()
                    return sentinel
                await queue.put(message)

        sentinel = object()
        handler_task = asyncio.create_task(self.app(scope, queue.get, send))
        asyncio.create_task(message_poller(sentinel, handler_task))
        try:
            return await handler_task
        except asyncio.CancelledError:
            print("Cancelling request due to disconnect")



import re
from typing import Optional

import py3langid as langid

from iso639 import Lang
from iso639.exceptions import InvalidLanguageValue

FLAIR_MODELS = {
    "en": "flair/ner-english-large",
    "es": "flair/ner-spanish-large",
    "de": "flair/ner-german-large",
    "nl": "flair/ner-dutch-large",
    "multi": "flair/ner-multi",                     # English, German, French, Spanish
    "multi-fast": "flair/ner-multi-fast",           # English, German, Dutch, Spanish
}

SPACY_MODELS = {
    "en": "en_core_web_sm",
    "es": "es_core_news_sm",
    "fr": "fr_core_news_sm",
    "de": "de_core_news_sm",
    "it": "it_core_news_sm"
}


def to_name(alpha2: str) -> str:
    if not alpha2:
        raise ValueError("Language code must be a non-empty string.")
    return Lang(alpha2).name


def to_alpha2(language_name: str) -> str:
    if not language_name:
        raise ValueError("Language name must be a non-empty string.")

    name = re.sub(r'\b\w+', lambda m: m.group(0).capitalize(), language_name)
    return Lang(name).pt1


def detect_language(text, min_confidence=None):
    detector = langid.langid.LanguageIdentifier.from_pickled_model(
        langid.langid.MODEL_FILE, norm_probs=True
    )
    detected_lang, confidence = detector.classify(text)
    if min_confidence and confidence < min_confidence:
        return None, confidence
    detected_lang = to_name(detected_lang)
    return detected_lang, confidence


def is_language_available(language: Optional[str], type: str) -> bool:
    if not language:
        return False

    try:
        lang_code = to_alpha2(language)

    except InvalidLanguageValue:
        lang_code = language

    match type:
        case "spacy":
            return SPACY_MODELS.get(lang_code, None) is not None

        case "flair":
            return FLAIR_MODELS.get(lang_code, None) is not None


def load_language_model(language, type):
    from flair.models import SequenceTagger
    from spacy_download import load_spacy

    match type:
        case "spacy":
            if is_language_available(language, "spacy"):
                model_name = SPACY_MODELS.get(to_alpha2(language), SPACY_MODELS["en"])
                return load_spacy(model_name)

            raise Exception(f"SpaCy model for language '{language}' is not available.")

        case "flair":
            if is_language_available(language, "flair"):
                model_name = FLAIR_MODELS.get(language, "flair/ner-multi")
                return SequenceTagger.load(model_name)

            raise Exception(f"Flair model for language '{language}' is not available.")



from enum import Enum
from .generics import Info, CombineField


class EntityInfo(Info):
    def __init__(self, unique_id, concept_name, fields=dict(), parents=list()):
        self.fields = fields
        self.concept_name = concept_name
        self.unique_id = unique_id
        self.parents = parents


class RelationshipInfo(Info):
    def __init__(self, concept_from, concept_to, name):
        self.concept_from = concept_from
        self.concept_to = concept_to
        self.name = name


class CombineFieldInfo(Info):
    def __init__(self, combiner, data):
        self.combiner = combiner
        self.data = data



from .generics import Match
from .signals import (
    AddGlobalConceptSignal,
    ChangeContextEntitySignal,
    ChangeContextSignal,
    SpawnChildrenEntitySignal,
)
import itertools

import logging

logger = logging.getLogger(__name__)


class LiteralMatch(Match):
    def __init__(self, fields, keep=False, to_combine_param=None, **params):
        super().__init__(
            **{**{"keep": keep, "to_combine_param": to_combine_param}, **params}
        )
        if not isinstance(fields, list):
            fields = [fields]
        self.fields = fields

    def load(self, parser):
        super().load(parser)

    def apply_match(self, value):
        if "keep" not in self.params or not self.params["keep"]:
            value.used = True
            logger.info("Setting " + value.key + " to used")
        param_name = None
        if "to_combine_param" in self.params and self.params["to_combine_param"]:
            param_name = self.params["to_combine_param"]
        return [{field: param_name for field in self.fields}], [
            {field: value.value for field in self.fields}
        ]


class EntityMatch(Match):
    def __init__(
        self,
        concept_name,
        parse_rules=None,
        inherits_rules_from_concepts=None,
        relationship_name=None,
        combine_rules=None,
        keep=True,
        push_context=False,
        **params
    ):
        super().__init__(**{**{"keep": keep, "push_context": push_context}, **params})
        self.concept_name = concept_name
        self.relationship_name = relationship_name
        if isinstance(inherits_rules_from_concepts, str):
            self.inherits_rules_from_concepts = [inherits_rules_from_concepts]
        else:
            self.inherits_rules_from_concepts = inherits_rules_from_concepts
        if not combine_rules:
            combine_rules = dict()
        self.combine_rules = combine_rules
        if not parse_rules:
            parse_rules = dict()
        self.parse_rules = parse_rules
        if not inherits_rules_from_concepts:
            inherits_rules_from_concepts = list()
        self.inherits_rules_from_concepts = inherits_rules_from_concepts

    def load(self, parser):
        super().load(parser)
        if self.parse_rules:
            for parse_rule, parse_match in self.parse_rules.items():
                parse_rule.load(self.parser)
                parse_match.load(self.parser)
        if self.combine_rules:
            for combine_rule in self.combine_rules.values():
                combine_rule.load(self.parser)

    def apply_match(self, value):
        if "keep" not in self.params or not self.params["keep"]:
            value.used = True
            logger.info("Setting " + value.key + " to used")
        new_entity = SpawnChildrenEntitySignal().send(
            self.parser,
            self.parser.curr_referred_entity,
            self.concept_name,
            self.relationship_name,
            dict(),
        )
        if "push_context" not in self.params or not self.params["push_context"]:
            push_context = False
        else:
            push_context = True
        ChangeContextEntitySignal(open_new_stack=not push_context).send(
            self.parser,
            new_entity,
            self.inherits_rules_from_concepts,
            self.parse_rules,
            self.combine_rules,
        )
        if push_context:
            ChangeContextSignal().send(self.parser, value)
        return [], []


class ConceptMatch(Match):
    def __init__(
        self,
        concept_name,
        concept_label=None,
        parse_rules=None,
        inherits_rules_from_concepts=None,
        relationship_name=None,
        combine_rules=None,
        keep=False,
        **params
    ):
        super().__init__(**{**{"keep": keep}, **params})
        self.concept_name = concept_name
        self.concept_label = concept_label
        self.relationship_name = relationship_name
        if isinstance(inherits_rules_from_concepts, str):
            self.inherits_rules_from_concepts = [inherits_rules_from_concepts]
        else:
            self.inherits_rules_from_concepts = inherits_rules_from_concepts
        self.parse_rules = parse_rules
        if not combine_rules:
            combine_rules = dict()
        self.combine_rules = combine_rules
        if not parse_rules:
            parse_rules = dict()
        self.parse_rules = parse_rules
        if not inherits_rules_from_concepts:
            inherits_rules_from_concepts = list()
        self.inherits_rules_from_concepts = inherits_rules_from_concepts

    def load(self, parser):
        super().load(parser)
        if self.parse_rules:
            for parse_rule, parse_match in self.parse_rules.items():
                parse_rule.load(self.parser)
                parse_match.load(self.parser)
        if self.combine_rules:
            for combine_rule in self.combine_rules.values():
                combine_rule.load(self.parser)
        AddGlobalConceptSignal(self).send(self.parser)

    def apply_match(self, value):
        if "keep" not in self.params or not self.params["keep"]:
            value.used = True
            logger.info("Setting " + value.key + " to used")
        new_entity = SpawnChildrenEntitySignal().send(
            self.parser,
            self.parser.curr_referred_entity,
            self.concept_name,
            self.relationship_name,
            dict(),
        )
        ChangeContextEntitySignal(open_new_stack=False).send(
            self.parser,
            new_entity,
            self.inherits_rules_from_concepts,
            self.parse_rules,
            self.combine_rules,
        )
        ChangeContextSignal().send(self.parser, value)
        return [], []


class ListExpandMatch(Match):
    class _int_iterator(object):
        def __init__(self):
            self.iterator = itertools.count()

        def __iter__(self):
            return self.iterator

        def __next__(self):
            return "_" + str(next(self.iterator))

    def __init__(self, prefix=None, labels_iter=_int_iterator, keep=False, **params):
        super().__init__(**{**{"keep": keep}, **params})
        self.labels_iter = labels_iter
        self.prefix = prefix

    def load(self, parser):
        super().load(parser)

    def apply_match(self, value):
        if "keep" not in self.params or not self.params["keep"]:
            value.used = True
            logger.info("Setting " + value.key + " to used")
        prefix = self.prefix
        if prefix is None:
            prefix = value.key
        new_value = dict()
        labels_iter = self.labels_iter()
        for i in range(len(value.value)):
            key = prefix + str(next(labels_iter))
            new_value[key] = value.value[i]
        value.value = new_value
        ChangeContextSignal().send(self.parser, value)
        return [], []


class DictExpandMatch(Match):
    def __init__(self, prefix="", keep=False, **params):
        super().__init__(**{**{"keep": keep}, **params})
        self.prefix = prefix

    def load(self, parser):
        super().load(parser)

    def apply_match(self, value):
        if "keep" not in self.params or not self.params["keep"]:
            value.used = True
            logger.info("Setting " + value.key + " to used")
        prefix = self.prefix
        if prefix is None:
            prefix = value.key
        new_value = dict()
        for k, v in value.value.items():
            key = prefix + k
            new_value[key] = v
        value.value = new_value
        ChangeContextSignal().send(self.parser, value)
        return [], []


class DiscardMatch(Match):
    def __init__(self, **params):
        super().__init__(**params)

    def load(self, parser):
        super().load(parser)

    def apply_match(self, value):
        value.used = True
        return [], []


class FunctionMatch(Match):
    def __init__(self, fn, keep=False, to_combine_param=None, **params):
        super().__init__(
            **{**{"keep": keep, "to_combine_param": to_combine_param}, **params}
        )
        self.fn = fn

    def load(self, parser):
        super().load(parser)

    def apply_match(self, value):
        param_name = None
        if "to_combine_param" in self.params and self.params["to_combine_param"]:
            param_name = self.params["to_combine_param"]
        func_return = self.fn(value.value)
        return_value = None
        if isinstance(func_return, (str, list)):
            inner_matcher = LiteralMatch(func_return, to_combine_param=param_name)
            inner_matcher.load(self.parser)
            return_value = inner_matcher.apply_match(value)
        elif isinstance(func_return, Match):
            inner_matcher = func_return
            inner_matcher.load(self.parser)
            return_value = inner_matcher.apply_match(value)
        elif isinstance(func_return, dict):
            return_value = [{key: param_name for key in func_return.keys()}], [
                func_return
            ]
        else:
            raise AttributeError("Invalid return type for FunctionMatch")
        if "keep" not in self.params or not self.params["keep"]:
            value.used = True
            logger.info("Setting " + value.key + " to used")
        return return_value


class ManyMatch(Match):
    def __init__(self, matches, keep=False, **params):
        super().__init__(**{**{"keep": keep}, **params})
        self.matches = matches

    def load(self, parser):
        super().load(parser)
        for match in self.matches:
            match.load(parser)

    def apply_match(self, value):
        return_list = list()
        value_original_used = value.used
        value_once_used = value.used
        for match in self.matches:
            return_list.append(match.apply_match(value))
            value_once_used = value_once_used or value.used
            value.used = value_original_used  # Reset usage until end of matches
        value.used = value.used or value_once_used
        if "keep" not in self.params or not self.params["keep"]:
            value.used = True
            logger.info("Setting " + value.key + " to used")
        return list(
            itertools.chain.from_iterable([elem[0] for elem in return_list])
        ), list(itertools.chain.from_iterable([elem[1] for elem in return_list]))



import re
import logging
from spacy_download import load_spacy
from flair.data import Sentence
from flair.models import SequenceTagger
from sumy.parsers.plaintext import PlaintextParser
from sumy.nlp.tokenizers import Tokenizer
from sumy.nlp.stemmers import Stemmer
from sumy.summarizers.lsa import LsaSummarizer
from sumy.summarizers.luhn import LuhnSummarizer
from sumy.utils import get_stop_words
import itertools
import numpy as np
import math
import nltk
from pydantic import BaseModel, ConfigDict, Field

logger = logging.getLogger(__name__)

SPACY_NER_MODELS = {
    "english": lambda: load_spacy(
        "en_core_web_sm",
        disable=["tok2vec", "tagger", "parser", "attribute_ruler", "lemmatizer"],
    )
}
FLAIR_NER_MODELS = {"english": lambda: SequenceTagger.load("flair/ner-english")}
REGEX_NER_MODELS = {
    "IP_ADDRESS": [
        r"\b(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)(?::(?:[0-9]|[1-9][0-9]{1,3}|[1-5][0-9]{4}|6[0-4][0-9]{3}|65[0-4][0-9]{2}|655[0-2][0-9]|6553[0-5]))?\b",
    ],
    "PHONE": r"(?:(?:\+(?:\d{1,3}[ .-]?)?(?:\(\d{1,3}\)[ .-]?)?)(?:\d{2,5}[ .-]?){1,3}|\d{2,5}[ .-]\d{2,5}(?:[ .-]\d{2,5}){0,2})\b",
    "EMAIL": r"\b[a-zA-Z0-9._%+-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]*[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]*[a-zA-Z0-9])?)+\b",
    "URL": r"\b(?:(?:https?|ftp|sftp|ftps|ssh|file|mailto|git|onion|ipfs|ipns):\/\/|www\.)(?:[a-z0-9](?:[a-z0-9-]{0,61}[a-z0-9])?\.)+[a-z]{2,}(?::\d+)?(?:\/(?:[-a-z0-9\/_.,~%+:@]|(?:%[0-9a-f]{2}))*)?(?:\?(?:[-a-z0-9\/_.,~%+:@=&]|(?:%[0-9a-f]{2}))*)?(?:#(?:[-a-z0-9\/_.,~%+:@=&]|(?:%[0-9a-f]{2}))*)?|(?:https?:\/\/)?[a-z2-7]{16,56}\.onion(?:\/(?:[-a-z0-9\/_.,~%+:@]|(?:%[0-9a-f]{2}))*)?(?:\?(?:[-a-z0-9\/_.,~%+:@=&]|(?:%[0-9a-f]{2}))*)?(?:#(?:[-a-z0-9\/_.,~%+:@=&]|(?:%[0-9a-f]{2}))*)\b",
}

BASE_TO_ONTONOTES_LABELMAP = {"PER": "PERSON"}
BASE_ALLOWED_LABELS = ["PERSON", "ORG", "LOC", "NORP", "GPE", "PRODUCT", "DATE", "PHONE", "IP_ADDRESS", "EMAIL", "URL"]


def _sumy__get_best_sentences(sentences, rating, *args, **kwargs):
    from operator import attrgetter
    from sumy.summarizers._summarizer import SentenceInfo

    rate = rating
    if isinstance(rating, dict):
        assert not args and not kwargs
        rate = lambda s: rating[s]
    infos = (SentenceInfo(s, o, rate(s, *args, **kwargs)) for o, s in enumerate(sentences))
    infos = sorted(infos, key=attrgetter("rating"), reverse=True)
    return tuple((i.sentence, i.rating, i.order) for i in infos)


def _sumy__luhn_call(summarizer, document):
    words = summarizer._get_significant_words(document.words)
    return _sumy__get_best_sentences(document.sentences, summarizer.rate_sentence, words)


def _sumy__lsa_call(summarizer, document):
    summarizer._ensure_dependecies_installed()
    dictionary = summarizer._create_dictionary(document)
    if not dictionary:
        return ()
    matrix = summarizer._create_matrix(document, dictionary)
    matrix = summarizer._compute_term_frequency(matrix)
    from numpy.linalg import svd as singular_value_decomposition

    u, sigma, v = singular_value_decomposition(matrix, full_matrices=False)
    ranks = iter(summarizer._compute_ranks(sigma, v))
    return _sumy__get_best_sentences(document.sentences, lambda s: next(ranks))


def get_nltk_tokenizer(language: str) -> Tokenizer:
    nltk.data.find("tokenizers/punkt")
    return Tokenizer(language)


class NERObject(BaseModel):
    name: str
    label: str
    score: float = 0.0
    start: int
    count: int
    context: str | None = None
    comentions: list[str] = Field(default_factory=list)
    model_config = ConfigDict(extra="allow")

    def __repr__(self):
        return f"NERObject(label={self.label},name={self.name})"


def postprocess_ner(entities: list[NERObject], whitelisted_labels=None, max_entities=None):
    if whitelisted_labels is not None:
        entities = [e for e in entities if e.label in whitelisted_labels]
    entities = sorted(entities, key=lambda x: x.name)
    final_entities = []
    for _, group in itertools.groupby(entities, key=lambda x: x.name):
        group = list(group)
        best_entity = max(group, key=lambda x: x.score * x.count)
        merged_data = {
            "name": best_entity.name,
            "label": best_entity.label,
            "score": best_entity.score,
            "context": best_entity.context,
            "count": sum(e.count for e in group),
            "start": best_entity.start,
        }
        all_fields = best_entity.model_fields.keys()
        for field in all_fields:
            if field in merged_data:
                continue
            values = [getattr(e, field, None) for e in group if getattr(e, field, None) is not None]
            if not values:
                continue
            if isinstance(values[0], list):
                merged_data[field] = list(set(itertools.chain.from_iterable(values or [])))
            else:
                merged_data[field] = getattr(best_entity, field, None)
        final_entities.append(NERObject(**merged_data))
    final_entities = sorted(final_entities, key=lambda x: x.score * x.count, reverse=True)
    if max_entities and len(final_entities) > max_entities:
        final_entities = final_entities[:max_entities]
    return final_entities


def compute_ner(
    language,
    sentences,
    spacy_model,
    flair_model=None,
    context_width=150,
    with_scores=True,
    with_comentions=True,
    with_context=True,
):
    sentence_starts = [0] + [len(s[0]) + 1 for s in sentences]
    del sentence_starts[-1]
    sentence_starts = list(np.cumsum(sentence_starts))
    text = "\n".join([s[0] for s in sentences])
    min_score = 1.0
    entities: list[NERObject] = []

    # FLAIR model (if not fast)
    if flair_model:
        input = [Sentence(sentence[0]) for sentence in sentences]
        flair_model.predict(input)
        output = [e for sentence in input for e in sentence.get_spans("ner")]
        flair_entities = [
            NERObject(
                name=entity.text,
                label=BASE_TO_ONTONOTES_LABELMAP.get(
                    entity.annotation_layers["ner"][0].value,
                    entity.annotation_layers["ner"][0].value,
                ),
                score=entity.score,
                start=sentence_starts[input.index(entity[0].sentence)] + entity[0].start_position,
                count=1,
            )
            for entity in output
        ]
        min_score = min([min_score] + [e.score for e in flair_entities])
        entities += flair_entities
        del flair_entities

    # REGEX model
    for label, regexes in REGEX_NER_MODELS.items():
        if not isinstance(regexes, list):
            regexes = [regexes]
        for regex in regexes:
            regex_entities = [
                NERObject(
                    name=match.group(),
                    label=label,
                    score=min_score - 0.5,
                    count=1,
                    start=match.start(),
                )
                for match in re.finditer(regex, text)
            ]
            entities += regex_entities
    min_score = min([min_score] + [e.score for e in regex_entities])

    # SPACY model
    chunks = []
    chunk_start_offsets = []
    current_chunk = []
    current_length = 0
    offset = 0
    for sentence, _ in sentences:
        sentence_len = len(sentence) + 1
        if sentence_len > spacy_model.max_length:
            truncated = sentence[: spacy_model.max_length - 1]
            chunks.append(truncated)
            chunk_start_offsets.append(offset)
            offset += sentence_len
            continue
        if current_length + sentence_len > spacy_model.max_length:
            chunks.append("\n".join(current_chunk))
            chunk_start_offsets.append(offset - current_length)
            current_chunk = []
            current_length = 0
        current_chunk.append(sentence)
        current_length += sentence_len
        offset += sentence_len
    if current_chunk:
        chunks.append("\n".join(current_chunk))
        chunk_start_offsets.append(offset - current_length)
    for i, chunk in enumerate(chunks):
        doc = spacy_model(chunk)
        chunk_offset = chunk_start_offsets[i]
        for entity in doc.ents:
            entities.append(
                NERObject(
                    name=entity.text,
                    label=BASE_TO_ONTONOTES_LABELMAP.get(entity.label_, entity.label_),
                    score=min_score - 0.5,
                    start=chunk_offset + entity.start_char,
                    count=1,
                )
            )

    # Reformatting for consistency
    if not entities:
        return []
    if with_scores:
        min_entity_score = min([e.score for e in entities])
        max_entity_score = max([e.score for e in entities])
        entity_score_range = 1 if min_entity_score == max_entity_score else (max_entity_score - min_entity_score)
        for e in entities:
            e.score = (e.score - min_entity_score) / entity_score_range
        scores = list(np.searchsorted(sentence_starts, [e.start + 1 for e in entities]))
        scores = [sentences[i - 1][1] for i in scores]
        scores = [scores[i] + 10 * entities[i].score for i in range(len(entities))]
        for i in range(len(entities)):
            entities[i].score = scores[i]
    else:
        for i in range(len(entities)):
            entities[i].score = 0.0
    if with_comentions:
        for i in range(len(entities)):
            entity = entities[i]
            comentions = [
                entities[j].name
                for j in range(len(entities))
                if j != i and abs(entities[j].start - entity.start) < math.ceil(context_width / 2)
            ]
            entities[i].comentions = comentions
    if with_context:
        for i in range(len(entities)):
            entity = entities[i]
            if entity.start >= 0 and entity.start < len(text):
                left = max(0, entity.start - math.floor(context_width / 2))
                right = min(len(text), entity.start + math.ceil(context_width / 2))
                context = ("[..]" if left > 0 else "") + text[left:right] + ("[..]" if right < len(text) else "")
                entities[i].context = context
    return entities


def _compress_by_frequency(text, language, target_chars):
    tokenizer = get_nltk_tokenizer(language)
    parser = PlaintextParser.from_string(text, tokenizer)
    sentences = parser.document.sentences
    if not sentences:
        return text
    current_size = len(text)
    if current_size <= target_chars:
        return text
    word_freq = {}
    for sentence in sentences:
        for word in sentence.words:
            word_str = str(word)
            normalized = word_str.lower()
            word_freq[normalized] = word_freq.get(normalized, 0) + 1
    sentence_scores = []
    for sentence in sentences:
        score = 0
        for word in sentence.words:
            try:
                word_str = word.value if hasattr(word, 'value') else str(word)
            except (AttributeError, TypeError):
                word_str = str(word)
            normalized = word_str.lower()
            score += word_freq.get(normalized, 0)
        sentence_scores.append((sentence._text, score, sentence))
    sentence_scores.sort(key=lambda x: x[1], reverse=True)
    selected = []
    current_chars = 0
    for sent_text, score, sent_obj in sentence_scores:
        if current_chars < target_chars:
            selected.append((sent_obj, sent_text))
            current_chars += len(sent_text) + 1
    selected.sort(key=lambda x: sentences.index(x[0]))
    result = " ".join([s[1] for s in selected])
    return result


def _run_single_summarizer(text, language, max_chars, method, with_scores=False):
    text_len = len(text)
    logger.debug(f"  Tokenizing {text_len:,} chars for {method} summarization...")
    tokenizer = get_nltk_tokenizer(language)
    stemmer = Stemmer(language)
    parser = PlaintextParser.from_string(text, tokenizer)
    if method == "luhn":
        summarizer = LuhnSummarizer(stemmer)
        summarizer.stop_words = get_stop_words(language)
        scored_sentences = iter(_sumy__luhn_call(summarizer, parser.document))
    elif method == "lsa":
        summarizer = LsaSummarizer(stemmer)
        summarizer.stop_words = get_stop_words(language)
        scored_sentences = iter(_sumy__lsa_call(summarizer, parser.document))
    else:
        raise ValueError(f"Unknown method: {method}. Use 'luhn' or 'lsa'")
    summary = []
    summary_chars = 0
    summary_chars_penultimate = 0
    sentence_count = 0
    while summary_chars < max_chars:
        try:
            next_sentence = next(scored_sentences)
            summary.append(next_sentence)
            summary_chars_penultimate = summary_chars
            summary_chars += len(" " + next_sentence[0]._text)
            sentence_count += 1
        except StopIteration:
            break
    summary = sorted(summary, key=lambda x: x[2])
    summary = [(sentence[0]._text, sentence[1]) for sentence in summary]
    if summary_chars > max_chars:
        final_chars = max_chars - summary_chars_penultimate
        summary[-1] = (
            summary[-1][0][:final_chars],
            summary[-1][1],
        )
    if not with_scores:
        summary = " ".join([s[0] for s in summary])
    else:
        min_score = min([s[1] for s in summary]) if summary else 0
        max_score = max([min_score] + [s[1] for s in summary])
        score_range = 1 if min_score == max_score else (max_score - min_score)
        summary = [(s[0], (s[1] - min_score) / score_range) for s in summary]
    final_size = len(summary) if isinstance(summary, str) else sum(len(s[0]) + 1 for s in summary)
    logger.debug(f"  Final summary size: {final_size:,} chars")
    return summary


def get_extractive_summary(text, language, max_chars, fast=False, with_scores=False):
    """
    Perform extractive summarization using bracket-based algorithm selection.
    
    Each summarizer handles a specific character range bracket:
    - FREQUENCY: [∞, 1M] - Fast O(n) algorithm for texts over 1M chars
    - LUHN/LSA: [1M, 0] - Quality algorithms for texts up to 1M chars
    Args:
        text: Input text to summarize
        language: Language for tokenization
        max_chars: Maximum characters in final summary
        fast: If True, use luhn (faster); if False, use lsa (more accurate) for <=1M texts
        with_scores: Return scores with sentences
    Returns:
        String text if with_scores=False, or list of (sentence, score) tuples if with_scores=True
    """
    text_len = len(text)
    current_text = text
    summarization_done = False
    if text_len > 1_000_000:
        stage_target = max(1_000_000, max_chars)
        method = "freqanalysis"
        logger.debug(f"Compressing with {method.upper()} from {text_len:,} chars to {stage_target:,}")
        current_text = _compress_by_frequency(current_text, language, stage_target)
        text_len = len(current_text)
        summarization_done = True
    if text_len <= 1_000_000:
        if max_chars < text_len:
            stage_target = max_chars
            final_method = "luhn" if fast else "lsa"
            logger.debug(f"Compressing with {final_method.upper()} from {text_len:,} chars to {stage_target:,}")
            result = _run_single_summarizer(current_text, language, stage_target, final_method, with_scores=with_scores)
            logger.debug(f"<<< STAGE 2 COMPLETE")
            return result
        else:
            logger.debug(f"\n⊘ Stage 2 skipped: max_chars {max_chars:,} ≥ current text {text_len:,}")         
            if summarization_done:
                if with_scores:
                    final_method = "luhn" if fast else "lsa"
                    result = _run_single_summarizer(current_text, language, text_len, final_method, with_scores=True)
                else:
                    result = current_text
                return result
            else:
                if with_scores:
                    final_method = "luhn" if fast else "lsa"
                    result = _run_single_summarizer(current_text, language, text_len, final_method, with_scores=True)
                else:
                    result = current_text
                return result
    else:
        if with_scores:
            final_method = "luhn" if fast else "lsa"
            result = _run_single_summarizer(current_text, language, max_chars, final_method, with_scores=True)
        else:
            result = current_text
        return result

def ner_pipe(
    text,
    language,
    spacy_model,
    flair_model=None,
    fast=False,
    compression_ratio="auto",
    with_scores=True,
    with_comentions=True,
    with_context=True,
):
    """
    Perform NER on text with optional extractive summarization.
    
    Uses progressive refinement for large texts:
    - Stage 1: If text >1M chars, use frequency-based scoring (O(n)) to reduce to ~100K
    - Stage 2: Use luhn or lsa for final high-quality selection
    
    Args:
        text: Input text
        language: Language code
        spacy_model: Loaded spacy model
        flair_model: Optional flair model for better accuracy
        fast: If True, use faster methods (luhn vs lsa for summarization)
        compression_ratio: Ratio to compress text with summarization (auto = dynamic based on size)
        with_scores: Include confidence scores
        with_comentions: Include co-mentioned entities
        with_context: Include context windows
    
    Returns:
        List of NERObject entities
    """
    if compression_ratio == "auto":
        compression_ratio = max(1.0, len(text) / 15000) if fast else 1.0
    sentences = get_extractive_summary(
        text, 
        language, 
        int(len(text) / compression_ratio), 
        fast=fast,
        with_scores=True
    )
    ner = compute_ner(language, sentences, spacy_model, flair_model, 150, with_scores, with_comentions, with_context)
    return ner


def get_ner_handler(language, fast=False):
    """
    Get a configured NER handler function.
    Args:
        language: Language code
        fast: If True, skip Flair model and use faster summarization
    Returns:
        A function that performs NER on text
    """
    try:
        get_nltk_tokenizer(language)
    except LookupError:
        language = "english"
    spacy_model = SPACY_NER_MODELS.get(language, SPACY_NER_MODELS["english"])()
    flair_model = None if fast else FLAIR_NER_MODELS.get(language, FLAIR_NER_MODELS["english"])()
    return lambda text, compression_ratio="auto", with_scores=True, with_comentions=True, with_context=True: ner_pipe(
        text, language, spacy_model, flair_model, fast, compression_ratio, with_scores, with_comentions, with_context
    )